@@ -1,4 +1,4 @@
-from mhkit.tidal import device 
 from mhkit.tidal import graphics
 from mhkit.tidal import io
 from mhkit.tidal import resource 
+from mhkit.tidal import performance
